﻿/*
 * Created by SharpDevelop.
 * User: munet
 * Date: 2006/01/08
 * Time: 19:14
 * 
 * To change this template use Tools | Options | Coding | Edit Standard Headers.
 */

using System;
using System.Text;
using System.Text.RegularExpressions;
using System.Collections;

namespace Mumed {

	/// <summary>純粋な数字データを格納・操作するクラス</summary>
	public class MNum: ICloneable {
		/// <summary>格納データタイプ</summary>
		public NumDef   type;
		/// <summary>符号 falseなら負 trueなら正</summary>
		public bool     sign;
		/// <summary>数字（分子）</summary>
		public ulong[]  n;
		/// <summary>数字（分母）</summary>
		public ulong[]  d;
		/// <summary>数字（小数）</summary>
		public double[] f;

        /// <summary>MNum 初期化コンストラクタ</summary>
		public MNum() {
            this.type = NumDef.Integer;
			this.n    = new ulong[1];
			this.n[0] = 0;
			this.d    = new ulong[1];
			this.d[0] = 1;
			this.f    = new double[1];
			this.f[0] = 0.0;
			this.sign = true;
		}

        /// <summary>MNum 初期化コンストラクタ←string</summary>
        public MNum (string str) : this() {
            // 正負判定
            Regex r_isMinus = new Regex(@"^-");
            this.sign = !(r_isMinus.IsMatch(str));

            // 小数判定
            Regex r_isDec = new Regex(@"^(-?)(([0-9]+)\.([0-9]+))$");
            string[] s_isDec = r_isDec.Split(str);

            if (s_isDec.Length == 6) {
                this.type = NumDef.Decimal;
                this.f[0] = Math.Abs(Double.Parse(s_isDec[2]));
                this.f[0] *= this.sign ? 1 : -1;
                return;
            }

            // 分数判定
            Regex r_isFla = new Regex(@"^(-?)([0-9]+)/([0-9]+)$");
            string[] s_isFla = r_isFla.Split(str);

            if (s_isFla.Length == 5) {
                this.type = NumDef.Flaction;
                this.n[0] = UInt64.Parse(s_isFla[2]);
                this.d[0] = UInt64.Parse(s_isFla[3]);
                this.Reduce();
                return;
            }

            // 整数判定
            Regex r_isInt = new Regex(@"^(-?)([0-9]+)$");
            string[] s_isInt = r_isInt.Split(str);

            if (s_isInt.Length >= 3) {
                this.type = NumDef.Integer;
                this.n[0] = UInt64.Parse(s_isInt[2]);
                return;

	            //StringBuilder sbb = new StringBuilder();
                //for (int i = 0; i < s_isInt.Length; i++)
                //    sbb.Append(i.ToString() + " / " + s_isInt[i] + "\n");
                //System.Windows.Forms.MessageBox.Show(sbb.ToString());
            }
		}

        /// <summary>Clone</summary>
        /// <returns>複製されたMCalcオブジェクト</returns>
        public object Clone() {
            MNum C = new MNum();
            C.type = this.type;
            C.sign = this.sign;
            C.n    = (ulong[])this.n.Clone();
            C.d    = (ulong[])this.d.Clone();
            C.f    = (double[])this.f.Clone();
            return C;
        }

        /// <summary>ToString</summary>
        /// <returns>String化されたMCalcデータ</returns>
		public override string ToString() {
			StringBuilder sb = new StringBuilder();

			switch(this.type) {
				case NumDef.Integer:
					if (!this.sign) sb.Append('-');
					sb.Append(this.n[0]);
					break;
				case NumDef.Flaction:
					if (!this.sign) sb.Append('-');
					sb.Append(this.n[0]);
					sb.Append("/");
					sb.Append(this.d[0]);
					break;
				case NumDef.Decimal:
					sb.Append(this.f[0]);
					break;
			}
			return sb.ToString();
		}
        
        /// <summary>ゼロ判定</summary>
        /// <returns>ゼロならTrue, 非ゼロならFalseを返す</returns>
        public bool isZero() {
        	return (this.type == NumDef.Integer && this.n[0] == 0);
        }

        /// <summary>１判定</summary>
        /// <returns>１ならTrue, 非１ならFalseを返す</returns>
        public bool isOne() {
        	return (this.type == NumDef.Integer && this.n[0] == 1);
        }

        /// <summary>正負判定</summary>
        /// <returns>正ならTrue, 負ならFalseを返す</returns>
        public bool isMinus() {
        	return !this.sign;
        }

        /// <summary>値の取得</summary>
        public double GetDouble() {
			switch(this.type) {
				case NumDef.Integer:
        			return (double)this.n[0] * ( this.sign ? 1 : -1 );
				case NumDef.Flaction:
					return (double)(this.n[0]) / (this.d[0]) * ( this.sign ? 1 : -1 );
				case NumDef.Decimal:
					return this.f[0];
			}
        	return 0.0;
        }

        /// <summary>分数ならばを約分をする</summary>
		public void Reduce(){
			if ( this.type == NumDef.Flaction ) {
				if ( this.n[0] == 0 ) {
					this.d[0] = 1;
                    this.type = NumDef.Integer;
                } else {
					ulong i_p   = 2;
					ulong i_max = Math.Max(this.n[0], this.d[0]);
                    i_max = Math.Max(i_max, 9);
					while( ( i_p * i_p ) <= i_max ) {
						if ( this.n[0] % i_p == 0 && this.d[0] % i_p == 0 ) {
							this.n[0] /= i_p;
							this.d[0] /= i_p;
						} else {
							i_p++;
						}
					}
					if ( this.d[0] == 1 ) this.type = NumDef.Integer;
				}
			}
		}
		/// <summary>逆数</summary>
		public void Inverse() {
			ulong i_tmp;
			switch(this.type) {
				case NumDef.Integer:
					i_tmp = this.n[0];
					this.type = NumDef.Flaction;
					this.n[0] = this.d[0];
					this.d[0] = i_tmp;
					break;
				case NumDef.Flaction:
					i_tmp = this.n[0];
					this.n[0] = this.d[0];
					this.d[0] = i_tmp;
					if ( this.d[0] == 1 ) this.type = NumDef.Integer;
					break;
				case NumDef.Decimal:
					this.f[0] = 1 / this.f[0];
					break;
			}
            if (this.d[0] == 0) throw new DivideByZeroException();
		}

		#region OperateTion Code
        /// <summary>足し算</summary>
        /// <param name="a">足し算するMNum</param>
		public void OperateAdd(MNum a) {
			double db = new double();
			switch(this.type){
				case NumDef.Integer:
					switch (a.type) {
						case NumDef.Integer:
							// 整数 ＋ 整数
							if ( this.sign != a.sign ) {
								long sub  = (long)this.n[0] - (long)a.n[0];
								this.n[0] = (ulong)Math.Abs(sub);
								if ( sub < 0 ) this.sign = !this.sign;
							} else {
								this.n[0] += a.n[0];
							}
							break;
						case NumDef.Flaction:
							// 整数 ＋ 分数
							if ( this.sign != a.sign ) {
								long sub  = (long)(this.n[0] * a.d[0]) - (long)a.n[0];
								this.n[0] = (ulong)Math.Abs(sub);
								this.d[0] = a.d[0];
								if ( sub < 0 ) this.sign = !this.sign;
							} else {
								this.n[0] *= a.d[0];
								this.n[0] += a.n[0];
								this.d[0]  = a.d[0];
							}
							this.type = NumDef.Flaction;
							break;
						case NumDef.Decimal:
							// 整数 ＋ 小数
							db = (double)((long)this.n[0] * ( this.sign ? 1 : -1 ));
							this.f[0] = db + a.f[0];
							this.sign = this.f[0] >= 0;
							this.type = NumDef.Decimal;
							break;
					}
					break;
				case NumDef.Flaction:
					switch (a.type) {
						case NumDef.Integer:
							// 分数 ＋ 整数
							if ( this.sign != a.sign ) {
								long sub  = (long)this.n[0] - (long)(a.n[0] * this.d[0]);
								this.n[0] = (ulong)Math.Abs(sub);
								if ( sub < 0 ) this.sign = !this.sign;
							} else {
								this.n[0] += a.n[0] * this.d[0];
							}
							break;
						case NumDef.Flaction:
							// 分数 ＋ 分数
							if ( this.sign != a.sign ) {
								long sub  = (long)(this.n[0] * a.d[0]) - (long)(a.n[0] * this.d[0]);
								this.n[0] = (ulong)Math.Abs(sub);
								this.d[0] = this.d[0] * a.d[0];
								if ( sub < 0 ) this.sign = !this.sign;
							} else {
								this.n[0] = this.n[0] * a.d[0] + a.n[0] * this.d[0];
								this.d[0] = this.d[0] * a.d[0];
							}
							this.Reduce();
							break;
						case NumDef.Decimal:
							// 分数 ＋ 小数
							db = (double)this.n[0] / (double)this.d[0] * ( this.sign ? 1 : -1 );
							this.f[0] = db + a.f[0];
							this.sign = this.f[0] >= 0;
							this.type = NumDef.Decimal;
							break;
					}
					break;
				case NumDef.Decimal:
					switch (a.type) {
						case NumDef.Integer:
							// 小数 ＋ 整数
							db = (double)((long)a.n[0] * ( a.sign ? 1 : -1 ));
							this.f[0] += db;
							this.sign = this.f[0] >= 0;
							break;
						case NumDef.Flaction:
							// 小数 ＋ 分数
							db = (double)a.n[0] / (double)a.d[0] * ( a.sign ? 1 : -1 );
							this.f[0] += db;
							this.sign = this.f[0] >= 0;
							break;
						case NumDef.Decimal:
							// 小数 ＋ 小数
							this.f[0] += a.f[0];
							this.sign = this.f[0] >= 0;
							break;
					}
					break;
			}
		}

        /// <summary>掛け算</summary>
        /// <param name="a">掛け算するMNum</param>
        public void OperateMul(MNum a) {
			double db;
			switch(this.type){
				case NumDef.Integer:
					switch (a.type) {
						case NumDef.Integer:
							// 整数 × 整数
							this.n[0] *= a.n[0];
							this.sign   ^= !a.sign;
							break;
						case NumDef.Flaction:
							// 整数 × 分数
							this.type = NumDef.Flaction;
							this.n[0] *= a.n[0];
							this.d[0] *= a.d[0];
							this.sign   ^= !a.sign;
							this.Reduce();
							break;
						case NumDef.Decimal:
							// 整数 × 小数
							db = (double)((long)this.n[0] * ( this.sign ? 1 : -1 ));
							this.f[0] = db * a.f[0];
							this.sign = this.f[0] >= 0;
							this.type = NumDef.Decimal;
							break;
					}
					break;
				case NumDef.Flaction:
					switch (a.type) {
						case NumDef.Integer:
							// 分数 × 整数
							this.n[0] *= a.n[0];
							this.sign   ^= !a.sign;
							this.Reduce();
							break;
						case NumDef.Flaction:
							// 分数 × 分数
							this.n[0] *= a.n[0];
							this.d[0] *= a.d[0];
							this.sign   ^= !a.sign;
							this.Reduce();
							break;
						case NumDef.Decimal:
							// 分数 × 小数
							db = (double)this.n[0] / (double)this.d[0] * ( this.sign ? 1 : -1 );
							this.f[0] = db * a.f[0];
							this.sign = this.f[0] >= 0;
							this.type = NumDef.Decimal;
							break;
					}
					break;
				case NumDef.Decimal:
					switch (a.type) {
						case NumDef.Integer:
							// 小数 × 整数
							db = (double)((long)a.n[0] * ( a.sign ? 1 : -1 ));
							this.f[0] *= db;
							this.sign = this.f[0] >= 0;
							break;
						case NumDef.Flaction:
							// 小数 × 分数
							db = (double)a.n[0] / (double)a.d[0] * ( a.sign ? 1 : -1 );
							this.f[0] *= db;
							this.sign = this.f[0] >= 0;
							break;
						case NumDef.Decimal:
							// 小数 × 小数
							this.f[0] *= a.f[0];
							this.sign = this.f[0] >= 0;
							break;
					}
					break;
			}
		}

        /// <summary>剰余</summary>
        /// <param name="a">割り算するMNum</param>
        public void OperateMod(MNum a) {
            double db;
            switch (this.type) {
                case NumDef.Integer:
                    switch (a.type) {
                        case NumDef.Integer:
                            // 整数 ÷ 整数
                            this.n[0] %= a.n[0];
                            this.sign ^= !a.sign;
                            break;
                        case NumDef.Flaction:
                            // 整数 ÷ 分数
                            this.type = NumDef.Flaction;
                            this.n[0] *= a.d[0];
                            this.n[0] %= a.n[0];
                            this.d[0] *= a.d[0];
                            this.sign ^= !a.sign;
                            this.Reduce();
                            break;
                        case NumDef.Decimal:
                            // 整数 ÷ 小数
                            db = (double)((long)this.n[0] * (this.sign ? 1 : -1));
                            this.f[0] = db % a.f[0];
                            this.sign = this.f[0] >= 0;
                            this.type = NumDef.Decimal;
                            break;
                    }
                    break;
                case NumDef.Flaction:
                    switch (a.type) {
                        case NumDef.Integer:
                            // 分数 ÷ 整数
                            a.n[0] *= this.d[0];
                            this.n[0] %= a.n[0];
                            this.sign ^= !a.sign;
                            this.Reduce();
                            break;
                        case NumDef.Flaction:
                            // 分数 ÷ 分数
                            this.n[0] *= a.d[0];
                            a.n[0] *= this.d[0];
                            this.n[0] %= a.n[0];
                            this.d[0] *= a.d[0];
                            this.sign ^= !a.sign;
                            this.Reduce();
                            break;
                        case NumDef.Decimal:
                            // 分数 ÷ 小数
                            db = (double)this.n[0] / (double)this.d[0] * (this.sign ? 1 : -1);
                            this.f[0] = db % a.f[0];
                            this.sign = this.f[0] >= 0;
                            this.type = NumDef.Decimal;
                            break;
                    }
                    break;
				case NumDef.Decimal:
                    switch (a.type) {
                        case NumDef.Integer:
                            // 小数 ÷ 整数
                            db = (double)((long)a.n[0] * (a.sign ? 1 : -1));
                            this.f[0] %= db;
                            this.sign = this.f[0] >= 0;
                            break;
                        case NumDef.Flaction:
                            // 小数 ÷ 分数
                            db = (double)a.n[0] / (double)a.d[0] * (a.sign ? 1 : -1);
                            this.f[0] %= db;
                            this.sign = this.f[0] >= 0;
                            break;
                        case NumDef.Decimal:
                            // 小数 ÷ 小数
                            this.f[0] %= a.f[0];
                            this.sign = this.f[0] >= 0;
                            break;
                    }
                    break;
            }
        }
        #endregion
        
        /// <summary>三角関数の適用</summary>
        /// <param name="k">適用する三角関数の種類</param>
        public void OperateTri(TriDef k) {
        	double db = this.GetDouble();
        	double res = 0.0;
			switch (k) {
				case TriDef.Sin:
					// 正弦
					res = Math.Sin(db);
					break;
				case TriDef.Cos:
					// 余弦
					res = Math.Cos(db);
					break;
				case TriDef.Tan:
					// 正接
					res = Math.Tan(db);
					break;
				case TriDef.Asin:
					// 正弦の逆関数
					if (( -1.0 <= db ) && ( db <= 1.0 )) res = Math.Asin(db);
					break;
				case TriDef.Acos:
					// 余弦の逆関数
					if (( -1.0 <= db ) && ( db <= 1.0 )) res = Math.Acos(db);
					break;
				case TriDef.Atan:
					// 正接の逆関数
					res = Math.Atan(db);
					break;
				case TriDef.Sinh:
					// 双曲正弦
					res = Math.Sinh(db);
					break;
				case TriDef.Cosh:
					// 双曲余弦
					res = Math.Cosh(db);
					break;
				case TriDef.Tanh:
					// 双曲正接
					res = Math.Tanh(db);
					break;
			}
        	if ( res.ToString().IndexOf(".") > 0 ) {
	        	this.f[0] = res;
				this.type = NumDef.Decimal;
        	} else {
        		this.n[0] = (ulong)Math.Abs(res);
        		this.d[0] = 1;
        		this.type = NumDef.Integer;
        	}
			this.sign = res >= 0;
        }

        /// <summary>対数の適用</summary>
        /// <param name="d">対数の底</param>
        public void OperateLog(double d) {
        	double db = this.GetDouble();
        	double res = Math.Log(db, d);
        	if ( res.ToString().IndexOf(".") > 0 ) {
	        	this.f[0] = res;
				this.type = NumDef.Decimal;
        	} else {
        		this.n[0] = (ulong)Math.Abs(res);
        		this.d[0] = 1;
        		this.type = NumDef.Integer;
        	}
			this.sign = res >= 0;
        }
        /// <summary>自然対数の適用</summary>
        public void OperateLog() {
        	this.OperateLog(Math.E);
        }
	}
 }
